"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

# -*- coding: utf-8 -*-
from datetime import datetime
from constants import CLUSTER_TIMEOUT
from constants import LATENCY_UPDATE_INTERVAL
from constants import MULTICLUSTER_UPDATE_INTERVAL
from utilities.config_manager import config_manager
import traceback
import eventlet
import sys

class BackgroundTask(object):

    def __init__(self, time_to_run=0):

        self._time_to_reiterate = time_to_run
        self._time_to_run = time_to_run

    def get_time_to_run(self):

        return self._time_to_run

    def decrement_time_to_run(self, decrementation_factor):

        self._time_to_run -= decrementation_factor

    def reset_time_to_run(self):

        self._time_to_run = self._time_to_reiterate

    def update_time_to_reiterate(self, time_to_reiterate):

        self._time_to_reiterate = time_to_reiterate


class StatisticsTask(BackgroundTask):

    def __init__(self, current_app):
        self._current_app = current_app
        super(StatisticsTask, self).__init__(config_manager.get_update_interval())

    def run(self, cluster_manager):
        clusters = cluster_manager.get_clusters()
        for cluster in list(clusters.values()):
            if cluster.is_paused():
                continue
            if cluster.is_secured():
                user_obj = cluster.get_highest_role_user()

                if user_obj:
                    cluster.update_statistics(user=user_obj.get_id(),
                                      password=user_obj.get_password())
            else:
                cluster.update_statistics()
        self.update_time_to_reiterate(config_manager.get_update_interval())


class LatencyTask(BackgroundTask):

    def __init__(self, time_to_run=0):
        super(LatencyTask, self).__init__(time_to_run)

    def run(self, cluster_manager):
        clusters = cluster_manager.get_clusters()
        for cluster in list(clusters.values()):
            if cluster.is_paused():
                continue
            if cluster.is_secured():
                user_obj = cluster.get_highest_role_user()
                if user_obj:
                    cluster.update_latencies(user=user_obj.get_id(),
                                                password=user_obj.get_password())
            else:
                cluster.update_latencies()


class RemoveUnusedClusters(BackgroundTask):

    def __init__(self, time_to_run=0):
        super(RemoveUnusedClusters, self).__init__(time_to_run)

    def run(self, cluster_manager):
        cluster_manager.remove_unused_clusters()


class MultiClusterTask(BackgroundTask):

    def __init__(self, time_to_run=0):
        super(MultiClusterTask, self).__init__(time_to_run)

    def run(self, session_manager):
        sessions = session_manager.get_all_sessions()
        for session in sessions.values():
            if 'multicluster' in session:
                multicluster_obj = session['multicluster']
                multicluster_obj.update_multicluster()

def perform_scheduled_tasks(cluster_manager, current_app, session_manager):

    cluster_timeout_interval = CLUSTER_TIMEOUT * 60
    time_to_remove_clusters = cluster_timeout_interval

    update_statistics = StatisticsTask(current_app)

    update_latency = LatencyTask(LATENCY_UPDATE_INTERVAL)

    remove_unused_clusters = RemoveUnusedClusters(time_to_remove_clusters)

    update_multicluster = MultiClusterTask(MULTICLUSTER_UPDATE_INTERVAL)

    list_of_tasks = [update_statistics, remove_unused_clusters,
                     update_latency, update_multicluster]

    list_of_tasks.sort(key=lambda x: x.get_time_to_run())
    sleep_time = list_of_tasks[0].get_time_to_run()
    eventlet.sleep(sleep_time)
    for task in list_of_tasks:
        task.decrement_time_to_run(sleep_time)

    while True:
        try:
            start_time = datetime.now()

            if list_of_tasks[0] == update_multicluster:
                list_of_tasks[0].run(session_manager)
            else:
                list_of_tasks[0].run(cluster_manager)

            end_time = datetime.now()

            list_of_tasks[0].reset_time_to_run()
            for index in range(1, len(list_of_tasks)):
                list_of_tasks[index].decrement_time_to_run((end_time-start_time).seconds + ((end_time-start_time).microseconds)/1000000)

            list_of_tasks.sort(key=lambda x: x.get_time_to_run())
            sleep_time = list_of_tasks[0].get_time_to_run()
            eventlet.sleep(sleep_time)
            for task in list_of_tasks:
                task.decrement_time_to_run(sleep_time)
        except Exception:
            sys.stderr.write("Error in update_statistics_of_all_clusters")
            sys.stderr.write((traceback.format_exc()))
            eventlet.sleep(10)

